# (c) 2020 Cluso99 (Ray Rodrick)
# License: MIT
#
# RR20200122 v001  commence
# RR20200123 v009  basics wkg - single block
# RR20200124 v010  use start/end comments to define a block (no longer each line)
#            v011  process each block
# RR20200125 v018  basics wkg
#            v019  permit " |ax" as valid chars in table ie add "x" to output special mask
#                  for intervening spaces, replace with "?" and don't report error
#            v020  allow start with "x" for non "x" columns (cannot use "x" in column "x")
#            v021  extend table columns to max_tables=104 but still only allows 52 characters a-z,A-Z
#            v022  get filename from commandline or user
#                  build "P2-skip-tables-022.exe" from "C:\PythonCode\pyinstaller --onefile P2-skip-tables-022.py"
# RR20200126 v023  32bit patterns, cater for ## (augx) 
#            v024
#            v025  change start '~nn and ' + + +
#                  change end   '~-- and ' - - -
#                  allow for two sets ## (augx)
#            v026  tweek (wkg)
progid = 'P2-skip-tables-026'

# scan thru' spin2 file and calculate skip bitmap for execf bytecodes

from time import time
from datetime import datetime, date, timedelta
import os
import sys

posn = 0

def main_process():
    global posn

    # path_name       = 'C:/P2/Spin2/'
    # fname_in        = 'source.spin2'
    path_name       = ''
    fname_out       = '_execf'

    #==============================================================================#
    # Get Parameters                                                               #
    #==============================================================================#
    params   = get_parameters(['filename'])                                         # get parameters (filename)
    fname_in = params[0]
    if '.' not in fname_in:
        fname_in = fname_in + '.spin2'


    #==============================================================================#
    # Read source, extract / process, write out all relevant code blocks           #
    #==============================================================================#
    fnamei    = path_name + fname_in
    lst       = fnamei.rsplit('.',1)
    fnameo    = lst[0] + fname_out + '.' + lst[1]

    lines_out = []
    lines_out.append("'' " + progid + ":  " + fnamei + " --> " + fnameo)            # add filenames to output file

    lines_in  = read_file(fnamei)                                                   # read source


    posn = 0
    while posn < len(lines_in):
        lines = get_block(lines_in)                                                 # return next code block
        for line in lines:                                                          # add source block as comment to output file
            lines_out.append("''" + line)
        lines = (process_block(lines))                                              # return processed block of code
        lines_out.extend(lines)                                                     # add processed block to output file
        lines_out.append("''")                                                      # add blank comment to output file

    write_file(fnameo, lines_out)                                                   # write out all processed blocks


#==============================================================================#
# Get next block of code                                                       #
#==============================================================================#
def get_block(lines_in):
    global posn
    lines = []
    found = False
    for i in range(posn, len(lines_in)):
        posn = i+1                                                                  # save next position
        line = lines_in[i]
        if not found:
            if (line[0:4] == "'~22") or (line[0:4] == "'~32"):                      # start of block?
                found = True                                                        #   y: set flag 
                lines.append(line)                                                  #   y: and copy
        else:
            lines.append(line)                                                      # copy
            if line[0:4] == "'~--":                                                 # end of block?
                break                                                               #   y: exit 'for loop
            elif "##" in line:                                                      # AUGx reqd?
                if "'" in line:
                    tmp = line.split("'")
                else:
                    tmp = [line, ""]
                for x in range(line.count("##")):
                    line = " nop <**augx placeholder**> ".ljust(len(tmp[0])) + "'" + tmp[1]
                    lines.append(line)                                              # insert <augx> line(s)
#    print(lines)
    return(lines)


#==============================================================================#
# Process a code block                                                         #
#==============================================================================#
def process_block(lines_in):
    lines     = []                                                                  # list of processed lines
    labels    = []                                                                  # list of instruction lables
    instrs    = []                                                                  # list of instructions (placeholder)
    tables    = []                                                                  # list of skip tables
    tblhdr    = []                                                                  # list of skip tables header character a-z,A-Z
    # max_tables = 104                                                                # max no of skip tables (52 but can have 52 blank columns)

    #==============================================================================#
    # Process each code line                                                       #
    #==============================================================================#
    for line in lines_in:
        line = line.replace('\t','    ')                                            # replace <tab> with 4 spaces
        if line.startswith("'~--") and ("' - - -" in line):                         # end block?
            pass
        elif line.startswith("'~") and ("' + + +" in line):                         # start block?
            bits = int(line[2:4])                                                   #   y: set skip mask bit size 22/32
            #==============================================================================#
            # Setup empty mask tables                                                      #
            #==============================================================================#
            tmp        = line[line.find("' + + +"):]                                # setup the table length for masks
            max_tables = 2 * (tmp.count(" +"))                                      # allow both " + + +"
            n  = tmp.count("+")                                                     #   and      "++++++"
            if (n > max_tables):
                max_tables = n
            for t in range(0,max_tables):                                           # max table entries in a table
                tables.append('')                                                   # clear out table list
                tblhdr.append('')                                                   # clear out table header character
            #==============================================================================#
        elif (len(line.strip()) == 0):                                              # ignore blank lines
            pass                                                                    
        elif (line.strip().find("'") < 1):                                          # ignore comment only lines
            pass                                                                    
        elif (line.find("'") > 0):                                                  # comment?
            #==============================================================================#
            # process valid line...                                                        #
            #==============================================================================#
            char0 = line[0]
            if char0.isalpha():                                                     # label? (first char alpha)
                label = line[:line.find(" ")]                                       # get the label
                instr = line[len(label):line.find("'")].strip()                     # get the instruction
            elif char0 == "'":                                                      # comment line?
                label = ''                                                          #   no label
                instr = ''                                                          #   no instruction
            else:                                                                   # no label
                label = ''                                                          #   no label
                instr = line[:line.find("'")].strip()                               #   get the instruction
            comment =  line[line.find("'")+1:]                                      # get the comment

#            print(label +' : ' + instr +' : ' + comment)

            if len(instr) == 0:
                lines.append('** missing instruction: ' + line)

            labels.append(label)                                                    # add the label to the list
            instrs.append(instr)                                                    # add the instruction to the list
            comment = comment.ljust(max_tables)                                     # make length>=?? for a-z,A-Z columns

            #==============================================================================#
            # create the table from the comment info...                                    #
            #==============================================================================#
            for t in range(0,max_tables):                                           # create table entries for this line
                ch = comment[t]                                                     # get a table entry char
                if ch.isalpha():                                                    # a-z,A-Z ?
                    if tblhdr[t] == '':                                             #   first time?
                        tblhdr[t] = ch                                              #      y: so copy
                        tables[t] = '0' + tables[t]                                 #      y: add '0' to respective table array
                    else:                                                           #
                        if tblhdr[t] == ch:                                         #      if no, validate
                            tables[t] = '0' + tables[t]                             #         y: add '0' to respective table array
                        elif ch == 'x':                                             #      special 'x' sequence?
                            tables[t] = 'x' + tables[t]                             #         y: add 'x' to respective table array
                        elif (tblhdr[t] == 'x') and (ch != 'x'):                    #      special case 'x?
                            tblhdr[t] = ch                                          #         y: replace hdr ch
                            tables[t] = tables[t].replace('0','x')                  #         y: replace previous '0' entries with 'x'
                            tables[t] = '0' + tables[t]                             #         y: add '0' to respective table array
                        else:                                                       #      else
                            tblhdr[t] = '?'                                         #         n: mark as invalid table
                            tables[t] = '?' + tables[t]                             #         n: add '?' to respective table array
                elif ch == '|':
                    tables[t] = '1' + tables[t]                                     # add '1' to respective table array
                elif ch == ' ':
                    tables[t] = ' ' + tables[t]                                     # add ' ' to respective table array
                else:
                    tblhdr[t] = '?'                                                 # invalid: mark as invalid table
                    tables[t] = '?' + tables[t]                                     # invalid: add '?' to respective table array
        else:
            pass                                                                    # ??? how to validate if no comment???

    #==============================================================================#
    # process all table entries for this block...                                  #
    #==============================================================================#
    for t in range(0,len(tables)):                                                  # copy valid entries... 
        if tblhdr[t].isalpha():
#            print(str(t) + ': '+ tblhdr[t] + ': ' + tables[t] +'<')
            line = process_entry(labels, tblhdr[t], tables[t], bits)
            lines.append(line)
        elif tblhdr[t] == '?':                                                      # terminate once invalid entry found ie end of list
            break

    return(lines)


#==============================================================================#
# process a valid entry...                                                     #
#==============================================================================#
def process_entry(labels, ch, mask, size):                                          # process valid entries
    n     = len(mask) - len(mask.rstrip())                                          # calculate index to correct label
    label = labels[n]
    txt   = label + '_skip_' + ch 
    if size > 22:
        line  = txt.ljust(24) + '= '                                                # mask > 22 so no label offset
        linex = ''                                                                  # or shift <<10
    else:
        line  = txt.ljust(24) + '= ' + label.ljust(16) + '| '
        linex = " << 10  '"

    msk = mask.strip()
    if len(msk) > size:                                                               # > 22 bits? OK if extras are all 0's             
        tmp = msk[:-size]
        tmp = tmp.replace('0','')
        if len(tmp) > 0:        
            line = '** mask > ' + str(size) + ': ' + line

    msk = msk.replace(' ','?')                                                      # replace illegal spaces with "?"
    for i in range(0,len(msk)):                                                     # check all valid "0/1/x/?" ?
        if (msk[i] in '01x?'):
            pass
        else:
            line = '** mask invalid: ' + line
            break   
    
    tmp = ''
    for i in range(n):                                                              # replace mask right spaces by underscores
        tmp = tmp + '_'
    if len(msk) > size:
        txt   = "%" + msk[:-size] + '_' + msk[-size:] + tmp + linex
    else:
        txt   = "%" + msk
        txt   = txt.rjust(size+1-n)  + tmp + linex

    line  = line + txt

    return(line)

#------------------------------------------------------------------------------#
def write_file(fname,lines):                    # write file from list
    print('i) Writing:    ' + fname)
    with open(fname, 'w', encoding='utf-8') as f:
        for line in lines:
            f.write(line + '\n')
        f.close()
#------------------------------------------------------------------------------#
def read_file(fname):                           # read file into list
    print('i) Reading:    ' + fname)
    with open(fname, 'r', encoding='utf-8') as f:
        line = f.read()
        lines = line.split('\n')                # split into lines
        f.close()
        return(lines)
#------------------------------------------------------------------------------#
def startprog(txt):                             # start program message
    print('i) Program <' + txt + '> begin... ' + datetime.now().strftime('%Y-%m-%d %H:%M:%S'))
    return(time())                              # start time
#------------------------------------------------------------------------------#
def endprog(starttime):                         # end program message
    print('i) Program end. (' + str(round((time() - starttime)/60,1)) +'m)  ' + datetime.now().strftime('%Y-%m-%d %H:%M:%S'))
#------------------------------------------------------------------------------#
def get_parameters(lst):                        # get parameters from command line or user)
    n = len(lst)
    if (len(sys.argv)-1) == n:                  # get params from command line?
        params = sys.argv[1:]
    else:                                       # get from user
        params = []
        for i in range(0, n):
            z = input (lst[i] + '? ')
            params.append(str(z))
    line = 'i) '
    for i in range(0, n):
        if i > 0:
            line = line + ', '
        line = line + str(lst[i]) +' = '+ str(params[i])
    print(line)
    return(params)
#------------------------------------------------------------------------------#

#------------------------------------------------------------------------------#
starttime = startprog(progid)
main_process()                              # go do it
endprog(starttime)
#------------------------------------------------------------------------------#
